"""coBib's configuration.

This file contains both, the actual implementation of the `Config` class, as well as the runtime
`config` object, which gets exposed on the module level as `cobib.config.config`.
Note, that this last link will not point to the correct location in the online documentation due to
the nature of the lower-level import.
"""

import configparser
import copy
import importlib.util
import io
import logging
import os
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Optional, TextIO, Union

from cobib.utils.rel_path import RelPath

LOGGER = logging.getLogger(__name__)

ANSI_COLORS = [
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
]


class LabelSuffix(Enum):
    """Suffixes to disambiguate Entry labels."""

    ALPHA = lambda count: chr(96 + count)
    CAPTIAL = lambda count: chr(64 + count)
    NUMERIC = lambda count: str(count)  # pylint: disable=unnecessary-lambda


class Config(Dict[str, Any]):
    """coBib's configuration class.

    This class wraps the `dict` type and exposes the dictionary keys as attributes to ease access
    (for both, getting and setting).
    Furthermore, nested attributes can be set directly without having to ensure that the parent
    attribute is already present.

    References:
        [Recursively access dict via attributes](https://stackoverflow.com/a/3031270)
    """

    XDG_CONFIG_FILE: str = "~/.config/cobib/config.py"
    """The XDG-based standard configuration location."""
    # TODO: remove legacy configuration support on 1.1.2022
    LEGACY_XDG_CONFIG_FILE: str = "~/.config/cobib/config.ini"
    """The *legacy* XDG-based standard configuration location.
    This will be removed on January 1st, 2020."""

    DEFAULTS: Dict[str, Any] = {
        "logging": {
            "logfile": "~/.cache/cobib/cobib.log",
            "version": "~/.cache/cobib/version",
        },
        "commands": {
            "edit": {
                "default_entry_type": "article",
                "editor": os.environ.get("EDITOR", "vim"),
            },
            "open": {
                "command": "xdg-open" if sys.platform.lower() == "linux" else "open",
            },
            "search": {
                "grep": "grep",
                "grep_args": [],
                "ignore_case": False,
            },
        },
        "database": {
            "file": "~/.local/share/cobib/literature.yaml",
            "format": {
                "label_default": "{label}",
                "label_suffix": ("_", LabelSuffix.ALPHA),
                "suppress_latex_warnings": True,
            },
            "git": False,
            "stringify": {
                "list_separator": {
                    "file": ", ",
                    "tags": ", ",
                    "url": ", ",
                },
            },
        },
        "events": {},
        "parsers": {
            "bibtex": {
                "ignore_non_standard_types": False,
            },
        },
        "tui": {
            "default_list_args": ["-l"],
            "prompt_before_quit": True,
            "reverse_order": True,
            "scroll_offset": 3,
            "colors": {
                "cursor_line_fg": "white",
                "cursor_line_bg": "cyan",
                "top_statusbar_fg": "black",
                "top_statusbar_bg": "yellow",
                "bottom_statusbar_fg": "black",
                "bottom_statusbar_bg": "yellow",
                "search_label_fg": "blue",
                "search_label_bg": "black",
                "search_query_fg": "red",
                "search_query_bg": "black",
                "popup_help_fg": "white",
                "popup_help_bg": "green",
                "popup_stdout_fg": "white",
                "popup_stdout_bg": "blue",
                "popup_stderr_fg": "white",
                "popup_stderr_bg": "red",
                "selection_fg": "white",
                "selection_bg": "magenta",
            },
            "key_bindings": {
                "prompt": ":",
                "search": "/",
                "help": "?",
                "add": "a",
                "delete": "d",
                "edit": "e",
                "filter": "f",
                "modify": "m",
                "open": "o",
                "quit": "q",
                "redo": "r",
                "sort": "s",
                "undo": "u",
                "select": "v",
                "wrap": "w",
                "export": "x",
                "show": "ENTER",
            },
        },
        "utils": {
            "file_downloader": {
                "default_location": "~/.local/share/cobib",
                "url_map": {},
            },
            "journal_abbreviations": [],
        },
    }
    """The default settings."""

    # pylint: disable=super-init-not-called
    def __init__(self, value: Optional[Dict[str, Any]] = None) -> None:
        """Initializer of the recursive, attribute-access, dict-like configuration object.

        The initializer does nothing when `None` is given.
        When a dict is given, all values are set as attributes.

        Args:
            value: a dictionary of settings.
        """
        if value is None:
            pass
        elif isinstance(value, dict):
            self.update(**value)
        else:
            raise TypeError("expected dict")

    def __setitem__(self, key: str, value: Any) -> None:
        """Sets a key, value pair in the object's dictionary.

        Args:
            key: the attributes' name.
            value: the attributes' value.
        """
        if isinstance(value, dict) and not isinstance(value, Config):
            value = Config(value)
        super().__setitem__(key, value)

    def __setattr__(self, key: str, value: Any) -> None:
        """Use `__setitem__` to set attributes unless it is a private (`__`) field.

        This is necessary in order to avoid a RecursionError during the pdoc generation.

        Args:
            key: the attributes' name.
            value: the attributes' value.
        """
        if key[0:2] == "__":
            super().__setattr__(key, value)
        else:
            self.__setitem__(key, value)

    MARKER = object()
    """A helper object for detecting the nested recursion-threshold."""

    EXCEPTIONAL_KEYS = {"events", "url_map"}
    """A set of exceptional keys which do not cause automatic item creation. This is required in
    order to support `dict`-like configuration options properly."""

    def __getitem__(self, key: str) -> Any:
        """Gets a key from the configuration object's dictionary.

        If the key is not present yet, it is automagically initialized with an empty configuration
        object to allow recursive attribute-setting.

        Args:
            key: the queried attributes' name.

        Returns:
            The value of the queried attribute.
        """
        found = self.get(key, Config.MARKER)
        if found is Config.MARKER and key not in Config.EXCEPTIONAL_KEYS:
            found = Config()
            super().__setitem__(key, found)
        return found

    def __getattr__(self, key: str) -> Any:
        """Use `__getitem__` to get attributes unless it is a private (`__`) field.

        This is necessary in order to avoid a RecursionError during the pdoc generation.

        Args:
            key: the queried attributes' name.

        Returns:
            The value of the queried attribute.
        """
        if key[0:2] == "__":
            return self.get(key)
        return self.__getitem__(key)

    def update(self, **kwargs) -> None:  # type: ignore
        """Updates the configuration with a dictionary of settings.

        This function ensures values are deepcopied, too.

        Args:
            kwargs: key, value pairs to be added to the configuration data.
        """
        for key, value in kwargs.items():
            self.__setitem__(key, copy.deepcopy(value))

    @staticmethod
    def load(configpath: Optional[Union[str, Path, TextIO, io.TextIOWrapper]] = None) -> None:
        """Loads another configuration object at runtime.

        WARNING: The new Python-like configuration allows essentially arbitrary Python code so it is
        the user's responsibility to treat this with care!

        Args:
            configpath: the path to the configuration.
        """
        if configpath is not None:
            if isinstance(configpath, (TextIO, io.TextIOWrapper)):
                configpath = configpath.name
        elif RelPath(Config.XDG_CONFIG_FILE).exists():
            configpath = RelPath(Config.XDG_CONFIG_FILE).path
        elif RelPath(Config.LEGACY_XDG_CONFIG_FILE).exists():  # pragma: no cover
            configpath = RelPath(Config.LEGACY_XDG_CONFIG_FILE).path  # pragma: no cover
        else:  # pragma: no cover
            return  # pragma: no cover
        LOGGER.info("Loading configuration from default location: %s", configpath)

        if RelPath(Config.LEGACY_XDG_CONFIG_FILE).exists():
            msg = (
                "The configuration mechanism of coBib underwent a major re-design for version 3.0! "
                "This means, that the old `INI`-style configuration is deprecated and will be "
                "fully removed on 1.1.2022. Instead, the configuration is now done through a "
                "Python file. For guidance on how to convert your existing configuration please "
                "consult the man-page or my blog post: "
                "https://mrossinek.gitlab.io/programming/cobibs-new-configuration/"
                "\nIf you have successfully migrated your configuration you should delete the old "
                "file in order to remove this warning message."
            )
            print("\x1b[1;37;41m#############\x1b[0m", file=sys.stderr)
            print("\x1b[1;37;41m## WARNING ##\x1b[0m", file=sys.stderr)
            print("\x1b[1;37;41m#############\x1b[0m", file=sys.stderr)
            LOGGER.warning(msg)

        spec = importlib.util.spec_from_file_location("config", configpath)
        if spec is None:
            LOGGER.warning(
                "The config at %s could not be interpreted as a Python module.", configpath
            )
            # attempt to load legacy INI configuration
            Config.load_legacy_config(configpath)
        else:
            cfg = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(cfg)  # type: ignore

        try:
            # validate config
            config.validate()
        except RuntimeError as exc:
            LOGGER.error(exc)
            sys.exit(1)

    @staticmethod
    def load_legacy_config(configpath: Union[str, Path, TextIO, io.TextIOWrapper]) -> None:
        # pylint: disable=too-many-branches,too-many-nested-blocks
        """Loads a legacy `INI`-style configuration file.

        WARNING: This functionality will be removed on January 1st, 2022!
        Users are being warned when using this configuration method.

        Args:
            configpath: the path to the configuration.
        """
        ini_conf = configparser.ConfigParser()
        # make option names case-sensitive:
        ini_conf.optionxform = str  # type: ignore
        ini_conf.read(configpath)

        def _ignore_option(section: str, field: str, value: Optional[Any] = None) -> None:
            if value is not None:  # pragma: no cover
                LOGGER.warning(  # pragma: no cover
                    "Ignoring unknown option for %s/%s = %s", section, field, value
                )
            else:  # pragma: no cover
                LOGGER.warning(  # pragma: no cover
                    "Ignoring unknown setting %s", f"{section}/{field}"
                )

        # We need to manually iterate all sections and fields because some settings need to be moved
        # and/or need to be converted to the correct Python types
        for section in ini_conf.sections():
            if section == "DATABASE":
                for field, value in dict(ini_conf[section]).items():
                    if field in ["file"]:
                        config.database[field] = value
                    elif field in ["git"]:
                        try:
                            config.database[field] = ini_conf[section].getboolean(field)
                        except ValueError as exc:  # pragma: no cover
                            LOGGER.error(exc)  # pragma: no cover
                            _ignore_option(section, field, value)  # pragma: no cover
                    elif field == "open":
                        config.commands.open.command = value
                    elif field == "grep":
                        config.commands.search.grep = value
                    elif field == "search_ignore_case":
                        config.commands.search.ignore_case = ini_conf[section].getboolean(field)
                    else:
                        _ignore_option(section, field)  # pragma: no cover
            elif section == "FORMAT":
                for field, value in dict(ini_conf[section]).items():
                    if field == "default_entry_type":
                        config.commands.edit.default_entry_type = value
                    elif field == "ignore_non_standard_types":
                        try:
                            config.parsers.bibtex[field] = ini_conf[section].getboolean(field)
                        except ValueError as exc:  # pragma: no cover
                            LOGGER.error(exc)  # pragma: no cover
                            _ignore_option(section, field, value)  # pragma: no cover
                    elif field == "month":
                        if value == "int":
                            config.database.format.month = int
                        elif value == "str":
                            config.database.format.month = str
                        else:
                            _ignore_option(section, field, value)  # pragma: no cover
                    else:
                        _ignore_option(section, field)  # pragma: no cover
            elif section == "TUI":
                for field, value in dict(ini_conf[section]).items():
                    if field == "default_list_args":
                        config.tui[field] = value.split(" ")
                    elif field in ["prompt_before_quit", "reverse_order"]:
                        try:
                            config.tui[field] = ini_conf[section].getboolean(field)
                        except ValueError as exc:  # pragma: no cover
                            LOGGER.error(exc)  # pragma: no cover
                            _ignore_option(section, field, value)  # pragma: no cover
                    elif field in ["scroll_offset"]:
                        try:
                            config.tui[field] = ini_conf[section].getint(field)
                        except ValueError as exc:  # pragma: no cover
                            LOGGER.error(exc)  # pragma: no cover
                            _ignore_option(section, field, value)  # pragma: no cover
                    else:
                        _ignore_option(section, field)  # pragma: no cover
            elif section == "COLORS":
                for field, value in dict(ini_conf[section]).items():
                    config.tui.colors[field] = value
            elif section == "KEY_BINDINGS":
                for field, value in dict(ini_conf[section]).items():
                    config.tui.key_bindings[field.lower()] = value
            else:
                LOGGER.warning("Ignoring unknown config section %s", section)  # pragma: no cover

    def validate(self) -> None:
        """Validates the configuration at runtime.

        Raises:
            RuntimeError when an invalid setting is encountered.
        """
        LOGGER.info("Validating the runtime configuration.")

        # LOGGING section
        LOGGER.debug("Validating the LOGGING configuration section.")
        self._assert(
            isinstance(self.logging.logfile, str), "config.logging.logfile should be a string."
        )
        self._assert(
            self.logging.version is None or isinstance(self.logging.version, str),
            "config.logging.version should be a string or `None`.",
        )

        # COMMANDS section
        LOGGER.debug("Validating the COMMANDS configuration section.")
        # COMMANDS.EDIT section
        LOGGER.debug("Validating the COMMANDS.EDIT configuration section.")
        self._assert(
            isinstance(self.commands.edit.default_entry_type, str),
            "config.commands.edit.default_entry_type should be a string.",
        )
        self._assert(
            isinstance(self.commands.edit.editor, str),
            "config.commands.edit.editor should be a string.",
        )
        # COMMANDS.OPEN section
        LOGGER.debug("Validating the COMMANDS.OPEN configuration section.")
        self._assert(
            isinstance(self.commands.open.command, str),
            "config.commands.open.command should be a string.",
        )
        # COMMANDS.SEARCH section
        LOGGER.debug("Validating the COMMANDS.SEARCH configuration section.")
        self._assert(
            isinstance(self.commands.search.grep, str),
            "config.commands.search.grep should be a string.",
        )
        self._assert(
            isinstance(self.commands.search.grep_args, list),
            "config.commands.search.grep_args should be a list.",
        )
        self._assert(
            isinstance(self.commands.search.ignore_case, bool),
            "config.commands.search.ignore_case should be a boolean.",
        )

        # DATABASE section
        self._assert(
            isinstance(self.database.file, str), "config.database.file should be a string."
        )
        self._assert(
            isinstance(self.database.git, bool), "config.database.git should be a boolean."
        )
        # DATABASE.FORMAT section
        if "month" in self["database"]["format"].keys():
            LOGGER.warning(
                "The config.database.format.month setting is deprecated as of version 3.1.0! "
                "Instead, coBib will store the month as a three-letter code which is a common "
                "format for which most citation styles include macros. See also "
                "https://www.bibtex.com/f/month-field/"
            )
        self._assert(
            isinstance(self.database.format.label_default, str),
            "config.database.format.label_default should be a string.",
        )
        self._assert(
            isinstance(self.database.format.label_suffix, tuple)
            and len(self.database.format.label_suffix) == 2,
            "config.database.format.label_suffix should be a tuple of length 2.",
        )
        self._assert(
            isinstance(self.database.format.label_suffix[0], str),
            "The first entry of config.database.format.label_suffix should be a string.",
        )
        self._assert(
            callable(self.database.format.label_suffix[1]),
            "The first entry of config.database.format.label_suffix should be a function.",
        )
        self._assert(
            isinstance(self.database.format.suppress_latex_warnings, bool),
            "config.database.format.suppress_latex_warnings should be a boolean.",
        )
        self._assert(
            isinstance(self.database.stringify.list_separator.file, str),
            "config.database.stringify.list_separator.file should be a string.",
        )
        self._assert(
            isinstance(self.database.stringify.list_separator.tags, str),
            "config.database.stringify.list_separator.tags should be a string.",
        )
        self._assert(
            isinstance(self.database.stringify.list_separator.url, str),
            "config.database.stringify.list_separator.url should be a string.",
        )

        # EVENTS section
        self._assert(isinstance(self.events, dict), "config.events should be a dict.")
        for event in self.events:
            self._assert(
                event.validate(),
                f"config.events.{event} did not pass its validation check.",
            )

        # PARSER section
        self._assert(
            isinstance(self.parsers.bibtex.ignore_non_standard_types, bool),
            "config.parsers.bibtex.ignore_non_standard_types should be a boolean.",
        )

        # TUI section
        self._assert(
            isinstance(self.tui.default_list_args, list),
            "config.tui.default_list_args should be a list.",
        )
        self._assert(
            isinstance(self.tui.prompt_before_quit, bool),
            "config.tui.prompt_before_quit should be a boolean.",
        )
        self._assert(
            isinstance(self.tui.reverse_order, bool),
            "config.tui.reverse_order should be a boolean.",
        )
        self._assert(
            isinstance(self.tui.scroll_offset, int),
            "config.tui.scroll_offset should be an integer.",
        )

        # TUI.COLORS section
        LOGGER.debug("Validating the TUI.COLORS configuration section.")
        for name in self.DEFAULTS["tui"]["colors"]:
            self._assert(
                name in self.tui.colors.keys(), f"Missing config.tui.colors.{name} specification!"
            )

        for name, color in self.tui.colors.items():
            if name not in self.DEFAULTS["tui"]["colors"].keys() and name not in ANSI_COLORS:
                LOGGER.warning("Ignoring unknown TUI color: %s.", name)
            self._assert(
                bool(
                    (color in ANSI_COLORS)
                    or (
                        len(color.strip("#")) == 6
                        and tuple(int(color.strip("#")[i : i + 2], 16) for i in (0, 2, 4))
                    )
                ),
                f"Unknown color specification: {color}",
            )

        # TUI.KEY_BINDINGS section
        LOGGER.debug("Validating the TUI.KEY_BINDINGS configuration section.")
        for command in self.DEFAULTS["tui"]["key_bindings"]:
            self._assert(
                command in self.tui.key_bindings.keys(),
                f"Missing config.tui.key_bindings.{command} key binding!",
            )
        for command, key in self.DEFAULTS["tui"]["key_bindings"].items():
            self._assert(
                isinstance(key, str), f"config.tui.key_bindings.{command} should be a string."
            )

        # UTILS section
        LOGGER.debug("Validating the UTILS configuration section.")
        self._assert(
            isinstance(self.utils.file_downloader.default_location, str),
            "config.utils.file_downloader.default_location should be a string.",
        )
        self._assert(
            isinstance(self.utils.file_downloader.url_map, dict),
            "config.utils.file_downloader.url_map should be a dict.",
        )
        for pattern, repl in self.utils.file_downloader.url_map.items():
            self._assert(
                isinstance(pattern, str) and isinstance(repl, str),
                "config.utils.file_downloader.url_map should be a dict[str, str].",
            )

        self._assert(
            isinstance(self.utils.journal_abbreviations, list),
            "config.utils.journal_abbreviations should be a list.",
        )
        for abbrev in self.utils.journal_abbreviations:
            self._assert(
                isinstance(abbrev, tuple),
                "config.utils.journal_abbreviations should be a list of tuples.",
            )

    @staticmethod
    def _assert(expression: bool, error: str) -> None:
        """Asserts the expression is True.

        Args:
            expression: the expression to assert.
            error: the message of the `RuntimeError` upon assertion failure.

        Raises:
            RuntimeError with the specified error string.
        """
        if not expression:
            raise RuntimeError(error)

    def defaults(self) -> None:
        """Resets the configuration to the default settings."""
        # pylint: disable=consider-using-dict-items
        for section in self.DEFAULTS:
            if section == "events":
                # manually reset events
                self["events"] = {}
                continue
            self[section].update(**self.DEFAULTS[section])

    def get_ansi_color(self, name: str) -> str:
        r"""Returns an ANSI color code for the named color.

        Appending `_fg` and `_bg` to the color name will yield the configured colors for the fore-
        and background property, respectively.
        The [ANSI color code](https://en.wikipedia.org/wiki/ANSI_escape_code#3-bit_and_4-bit) can
        then be constructed using the formula `\x1b[{FG};{BG}m`.

        Args:
            name: a named color as specified in the configuration *excluding* the `_fg` or `_bg`
                  suffix.

        Returns:
            A string representing the foreground and background ANSI color code.
        """
        fg_color = 30 + ANSI_COLORS.index(self.tui.colors.get(name + "_fg"))
        bg_color = 40 + ANSI_COLORS.index(self.tui.colors.get(name + "_bg"))

        return f"\x1b[{fg_color};{bg_color}m"


config: Config = Config()
"""This is the runtime configuration object. It is exposed on the module level via:
```python
from cobib.config import config
```
"""
config.defaults()



"""Tests for coBib's TUI."""
# pylint: disable=no-self-use,unused-argument,too-many-public-methods

import copy
import tempfile
from itertools import zip_longest
from pathlib import PurePath
from signal import SIGWINCH
from typing import Any, Dict, Generator, List, Set, Union

import pytest

from cobib.config import config
from cobib.database import Database
from cobib.tui import TUI, Frame, TextBuffer
from cobib.tui.state import STATE, Mode

from .. import get_resource
from ..cmdline_test import CmdLineTest
from .mock_curses import MockCursesPad
from .tui_test import TUITest


class TestTUI(CmdLineTest, TUITest):
    """Tests for coBib's TUI.

    This class derives the `tests.tui.tui_test.TUITest` as well as the
    `tests.cmdline_test.CmdLineTest` because it also ensures that the TUI can be started from the
    command-line interface.
    """

    @staticmethod
    @pytest.fixture(autouse=True)
    def setup() -> Generator[Any, None, None]:
        """Setup.

        This fixture is automatically enabled for all tests in this class.

        Yields:
            This method yields control to the actual test after which it will tear down the setup.
        """
        # pylint: disable=attribute-defined-outside-init
        config.load(get_resource("debug.py"))
        original_keydict = copy.deepcopy(TUI.KEYDICT)
        yield
        # clean up config
        config.defaults()
        STATE.reset()
        TUI.KEYDICT = copy.deepcopy(original_keydict)

    def test_colors(self, patch_curses: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.tui.TUI.colors`.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
        """
        TUI.colors()
        assert TUI.ANSI_MAP == {
            "\x1b[30;43m": 2,
            "\x1b[34;40m": 3,
            "\x1b[31;40m": 4,
            "\x1b[37;46m": 5,
            "\x1b[37;42m": 6,
            "\x1b[37;44m": 7,
            "\x1b[37;41m": 8,
            "\x1b[37;45m": 9,
        }
        expected_log = [
            ("TUITest", 10, "start_color"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 1 for top_statusbar"),
            ("TUITest", 10, "init_pair: (1, 0, 3)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for top_statusbar"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 2 for bottom_statusbar"),
            ("TUITest", 10, "init_pair: (2, 0, 3)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for bottom_statusbar"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 3 for search_label"),
            ("TUITest", 10, "init_pair: (3, 4, 0)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for search_label"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 4 for search_query"),
            ("TUITest", 10, "init_pair: (4, 1, 0)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for search_query"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 5 for cursor_line"),
            ("TUITest", 10, "init_pair: (5, 7, 6)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for cursor_line"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 6 for popup_help"),
            ("TUITest", 10, "init_pair: (6, 7, 2)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for popup_help"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 7 for popup_stdout"),
            ("TUITest", 10, "init_pair: (7, 7, 4)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for popup_stdout"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 8 for popup_stderr"),
            ("TUITest", 10, "init_pair: (8, 7, 1)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for popup_stderr"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 9 for selection"),
            ("TUITest", 10, "init_pair: (9, 7, 5)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for selection"),
        ]
        assert caplog.record_tuples == expected_log

    def test_config_color(self, patch_curses: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.tui.TUI.colors` when setting a non-default color value.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
        """
        config.tui.colors.selection_fg = "red"
        TUI.colors()
        assert ("TUITest", 10, "init_pair: (9, 1, 5)") in caplog.record_tuples
        assert ("cobib.tui.tui", 10, "Adding ANSI color code for selection") in caplog.record_tuples

    @pytest.mark.parametrize(
        ["can_change_color"],
        [
            [False],
            [True],
        ],
    )
    def test_config_rgb_color(
        self,
        patch_curses: Any,
        caplog: pytest.LogCaptureFixture,
        monkeypatch: pytest.MonkeyPatch,
        can_change_color: bool,
    ) -> None:
        """Test overwriting the RGB color value.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
            can_change_color: whether `curses.can_change_color` is enabled.
        """
        monkeypatch.setattr("curses.can_change_color", lambda: can_change_color)
        config.tui.colors.white = "#AA0000"
        TUI.colors()
        if not can_change_color:
            assert (
                "cobib.tui.tui",
                30,
                "Curses cannot change the default colors. Skipping color setup.",
            ) in caplog.record_tuples
        else:
            assert ("TUITest", 10, "init_color: (7, 666, 0, 0)") in caplog.record_tuples

    def test_config_unknown_color(
        self, patch_curses: Any, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test that setting an unknown color logs a warning.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
        """
        config.tui.colors.dummy_fg = "white"
        TUI.colors()
        assert (
            "cobib.tui.tui",
            30,
            "Detected unknown TUI color name specification: dummy",
        ) in caplog.record_tuples

    def test_bind_keys(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.tui.TUI.bind_keys`.

        Args:
            caplog: the built-in pytest fixture.
        """
        TUI.bind_keys()
        assert TUI.KEYDICT == {
            258: ("y", 1),
            259: ("y", -1),
            338: ("y", 20),
            339: ("y", -20),
            106: ("y", 1),
            107: ("y", -1),
            103: ("y", "g"),
            71: ("y", "G"),
            2: ("y", -20),
            4: ("y", 10),
            6: ("y", 20),
            21: ("y", -10),
            260: ("x", -1),
            261: ("x", 1),
            104: ("x", -1),
            108: ("x", 1),
            48: ("x", 0),
            36: ("x", "$"),
            58: "Prompt",
            47: "Search",
            63: "Help",
            97: "Add",
            100: "Delete",
            101: "Edit",
            102: "Filter",
            109: "Modify",
            111: "Open",
            113: "Quit",
            114: "Redo",
            115: "Sort",
            117: "Undo",
            118: "Select",
            119: "Wrap",
            120: "Export",
            10: "Show",
            13: "Show",
        }
        expected_log = [
            ("cobib.tui.tui", 20, "Binding key : to the Prompt command."),
            ("cobib.tui.tui", 20, "Binding key / to the Search command."),
            ("cobib.tui.tui", 20, "Binding key ? to the Help command."),
            ("cobib.tui.tui", 20, "Binding key a to the Add command."),
            ("cobib.tui.tui", 20, "Binding key d to the Delete command."),
            ("cobib.tui.tui", 20, "Binding key e to the Edit command."),
            ("cobib.tui.tui", 20, "Binding key f to the Filter command."),
            ("cobib.tui.tui", 20, "Binding key m to the Modify command."),
            ("cobib.tui.tui", 20, "Binding key o to the Open command."),
            ("cobib.tui.tui", 20, "Binding key q to the Quit command."),
            ("cobib.tui.tui", 20, "Binding key r to the Redo command."),
            ("cobib.tui.tui", 20, "Binding key s to the Sort command."),
            ("cobib.tui.tui", 20, "Binding key u to the Undo command."),
            ("cobib.tui.tui", 20, "Binding key v to the Select command."),
            ("cobib.tui.tui", 20, "Binding key w to the Wrap command."),
            ("cobib.tui.tui", 20, "Binding key x to the Export command."),
            ("cobib.tui.tui", 20, "Binding key ENTER to the Show command."),
        ]
        assert caplog.record_tuples == expected_log

    def test_config_bind_key(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.tui.TUI.bind_keys` when binding a non-default key.

        Args:
            caplog: the built-in pytest fixture.
        """
        config.tui.key_bindings.prompt = "p"
        TUI.bind_keys()
        assert ord(":") not in TUI.KEYDICT.keys()
        assert ord("p") in TUI.KEYDICT and TUI.KEYDICT[ord("p")] == "Prompt"
        assert ("cobib.tui.tui", 20, "Binding key p to the Prompt command.") in caplog.record_tuples

    def test_config_unknown_command(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test that binding an unknown command logs a warning.

        Args:
            caplog: the built-in pytest fixture.
        """
        config.tui.key_bindings.dummy = "p"
        TUI.bind_keys()
        assert (
            "cobib.tui.tui",
            30,
            'Unknown command "Dummy". Ignoring key binding.',
        ) in caplog.record_tuples

    def test_infoline(self) -> None:
        """Test `cobib.tui.tui.TUI.infoline`."""
        infoline = TUI.infoline()
        assert (
            infoline
            == "a:Add d:Delete e:Edit x:Export f:Filter ?:Help m:Modify o:Open ::Prompt q:Quit "
            "r:Redo /:Search v:Select ENTER:Show s:Sort u:Undo w:Wrap"
        )

    @pytest.mark.parametrize(
        ["attr"],
        [
            [0],
            [1],
        ],
    )
    def test_statusbar(self, attr: int, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.tui.TUI.statusbar`.

        Args:
            attr: the attribute number to use for the printed text.
            caplog: the built-in pytest fixture.
        """
        pad = MockCursesPad()
        text = "Test statusbar text"
        TUI.statusbar(pad, text, attr)
        assert pad.lines == [text]
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "getmaxyx"),
            ("MockCursesPad", 10, f"addnstr: 0 0 Test statusbar text -1 {attr}"),
            ("MockCursesPad", 10, "refresh: None None None None None None"),
        ]
        assert caplog.record_tuples == expected_log

    def test_init(self, patch_curses: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.tui.TUI.__init__`.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
        """
        stdscr = MockCursesPad()
        stdscr.size = (24, 80)
        tui = TUI(stdscr, debug=True)
        assert tui.stdscr == stdscr
        assert tui.width == 80
        assert tui.height == 24
        assert tui.STATE == STATE
        assert tui.prompt_before_quit is True
        assert tui.selection == set()
        assert isinstance(tui.topbar, MockCursesPad)
        assert tui.topbar.lines[0].startswith("coBib")
        assert tui.topbar.lines[0].endswith("3 Entries")
        assert isinstance(tui.botbar, MockCursesPad)
        assert (
            tui.botbar.lines[0]
            == "a:Add d:Delete e:Edit x:Export f:Filter ?:Help m:Modify o:Open ::Prompt q:Quit "
            "r:Redo /:Search v:Select ENTER:Show s:Sort u:Undo w:Wrap"
        )
        assert isinstance(tui.prompt, MockCursesPad)
        assert isinstance(tui.viewport, Frame)
        assert tui.viewport.width == 80
        assert tui.viewport.height == 21
        expected_lines = [
            "knuthwebsite    Knuth: Computers and Typesetting",
            "latexcompanion  The \\LaTeX\\ Companion",
            'einstein        Zur Elektrodynamik bewegter K{\\"o}rper',
        ]
        for line, truth in zip_longest(expected_lines, tui.viewport.buffer.lines):
            assert line == truth.strip()
        expected_log = [
            ("cobib.tui.tui", 20, "Initializing TUI."),
            ("TUITest", 10, "curs_set: (0,)"),
            ("cobib.tui.tui", 10, "stdscr size determined to be 80x24"),
            ("cobib.tui.tui", 10, "Initializing colors."),
            ("TUITest", 10, "use_default_colors"),
            ("TUITest", 10, "start_color"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 1 for top_statusbar"),
            ("TUITest", 10, "init_pair: (1, 0, 3)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for top_statusbar"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 2 for bottom_statusbar"),
            ("TUITest", 10, "init_pair: (2, 0, 3)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for bottom_statusbar"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 3 for search_label"),
            ("TUITest", 10, "init_pair: (3, 4, 0)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for search_label"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 4 for search_query"),
            ("TUITest", 10, "init_pair: (4, 1, 0)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for search_query"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 5 for cursor_line"),
            ("TUITest", 10, "init_pair: (5, 7, 6)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for cursor_line"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 6 for popup_help"),
            ("TUITest", 10, "init_pair: (6, 7, 2)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for popup_help"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 7 for popup_stdout"),
            ("TUITest", 10, "init_pair: (7, 7, 4)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for popup_stdout"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 8 for popup_stderr"),
            ("TUITest", 10, "init_pair: (8, 7, 1)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for popup_stderr"),
            ("cobib.tui.tui", 10, "Initiliazing color pair 9 for selection"),
            ("TUITest", 10, "init_pair: (9, 7, 5)"),
            ("cobib.tui.tui", 10, "Adding ANSI color code for selection"),
            ("cobib.tui.tui", 10, "Initializing key bindings."),
            ("cobib.tui.tui", 20, "Binding key : to the Prompt command."),
            ("cobib.tui.tui", 20, "Binding key / to the Search command."),
            ("cobib.tui.tui", 20, "Binding key ? to the Help command."),
            ("cobib.tui.tui", 20, "Binding key a to the Add command."),
            ("cobib.tui.tui", 20, "Binding key d to the Delete command."),
            ("cobib.tui.tui", 20, "Binding key e to the Edit command."),
            ("cobib.tui.tui", 20, "Binding key f to the Filter command."),
            ("cobib.tui.tui", 20, "Binding key m to the Modify command."),
            ("cobib.tui.tui", 20, "Binding key o to the Open command."),
            ("cobib.tui.tui", 20, "Binding key q to the Quit command."),
            ("cobib.tui.tui", 20, "Binding key r to the Redo command."),
            ("cobib.tui.tui", 20, "Binding key s to the Sort command."),
            ("cobib.tui.tui", 20, "Binding key u to the Undo command."),
            ("cobib.tui.tui", 20, "Binding key v to the Select command."),
            ("cobib.tui.tui", 20, "Binding key w to the Wrap command."),
            ("cobib.tui.tui", 20, "Binding key x to the Export command."),
            ("cobib.tui.tui", 20, "Binding key ENTER to the Show command."),
            ("cobib.tui.tui", 10, "Initializing global State"),
            ("cobib.tui.tui", 10, "Populating top status bar."),
            ("cobib.tui.tui", 10, "Populating bottom status bar."),
            ("cobib.tui.tui", 10, "Initializing viewport with Frame"),
            ("cobib.tui.tui", 10, "Populating viewport buffer."),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] in ("cobib.tui.tui", "TUITest")
        ] == expected_log

    def test_resize(self, patch_curses: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.tui.TUI.resize_handler`.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
        """
        stdscr = MockCursesPad()
        stdscr.size = (24, 80)
        tui = TUI(stdscr, debug=True)
        caplog.clear()

        tui.height, tui.width = (12, 70)
        tui.resize_handler(None, None)
        assert tui.width == 70
        assert tui.height == 12
        assert tui.topbar.size[1] == 70  # type: ignore
        assert tui.botbar.size[1] == 70  # type: ignore
        assert tui.prompt.size[1] == 70  # type: ignore
        expected_log = [
            ("TUITest", 10, "resize_term"),
            ("MockCursesPad", 10, "clear"),
            ("MockCursesPad", 10, "refresh: None None None None None None"),
            ("MockCursesPad", 10, "resize: 1 70"),
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "getmaxyx"),
            ("MockCursesPad", 10, "addnstr: 0 0 coBib VERSION - 3 Entries 69 0"),  # will be skipped
            ("MockCursesPad", 10, "refresh: None None None None None None"),
            ("MockCursesPad", 10, "refresh: None None None None None None"),
            ("MockCursesPad", 10, "resize: 1 70"),
            ("MockCursesPad", 10, "mvwin: 10 0"),
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "getmaxyx"),
            (
                "MockCursesPad",
                10,
                "addnstr: 0 0 a:Add d:Delete e:Edit x:Export f:Filter ?:Help m:Modify o:Open "
                "::Prompt q:Quit r:Redo /:Search v:Select ENTER:Show s:Sort u:Undo w:Wrap 69 0",
            ),
            ("MockCursesPad", 10, "refresh: None None None None None None"),
            ("MockCursesPad", 10, "refresh: None None None None None None"),
            ("MockCursesPad", 10, "resize: 1 70"),
            ("MockCursesPad", 10, "refresh: 0 0 11 0 12 69"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 9 69"),
        ]
        for log, truth in zip(
            expected_log,
            [
                record
                for record in caplog.record_tuples
                if record[0] in ("MockCursesPad", "TUITest")
            ],
        ):
            assert log[0] == truth[0]
            assert log[1] == truth[1]
            if truth[2].startswith("addnstr: 0 0 coBib v"):
                # skip version-containing log
                continue
            assert log[2] == truth[2]

    def test_resize_live(self) -> None:
        """Test `cobib.tui.tui.TUI.resize_handler` while the TUI is actually running."""

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_screen = [
                "knuthwebsite    Knuth: Computers and Typesett",
                r"latexcompanion  The \LaTeX\ Companion",
                "einstein        Zur Elektrodynamik bewegter K",
                "",
                "",
                "",
                "",
                "a:Add d:Delete e:Edit x:Export f:Filter ?:He",
            ]
            for line, truth in zip(expected_screen, screen.display[1:]):
                assert line == truth.strip()

            expected_log = [
                ("cobib.tui.tui", 10, "Handling resize event."),
                ("cobib.tui.tui", 10, "New stdscr dimension determined to be 45x10"),
            ]
            assert all(log in logs for log in expected_log)

        self.run_tui([SIGWINCH], assertion, {})

    def test_help(self, patch_curses: Any, caplog: pytest.LogCaptureFixture) -> None:
        # pylint: disable=consider-using-f-string
        """Test `cobib.tui.tui.TUI.help`.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
        """
        stdscr = MockCursesPad()
        stdscr.size = (24, 80)
        tui = TUI(stdscr, debug=True)
        caplog.clear()

        tui.help()
        expected_log = [
            ("cobib.tui.tui", 10, "Help command triggered."),
            ("cobib.tui.tui", 10, "Generating help text."),
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 22 80"),
            ("MockCursesPad", 10, "resize: 21 80"),
            ("MockCursesPad", 10, "addstr: 1 1                              coBib TUI Help"),
            ("MockCursesPad", 10, "addstr: 2 1   Key    Command  Description"),
            ("MockCursesPad", 10, "bkgd:   (6,)"),
            ("MockCursesPad", 10, "box"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 22 80"),
            ("MockCursesPad", 10, "getch"),
            ("MockCursesPad", 10, "clear"),
            ("cobib.tui.tui", 10, "Handling resize event."),
        ]
        inv_keys = {}
        for key, cmd in TUI.KEYDICT.items():
            if cmd in TUI.HELP_DICT:
                inv_keys[cmd] = "ENTER" if key in (10, 13) else chr(key)
        for idx, (cmd, desc) in enumerate(TUI.HELP_DICT.items()):
            expected_log.insert(
                -6,
                (
                    "MockCursesPad",
                    10,
                    f"addstr: {3+idx} 1 "
                    + "{:^8} {:<8} {}".format(
                        "[" + config.tui.key_bindings[cmd.lower()] + "]", cmd + ":", desc
                    ),
                ),
            )
        for log, truth in zip(
            expected_log,
            [
                record
                for record in caplog.record_tuples
                if record[0] in ("MockCursesPad", "cobib.tui.tui")
            ],
        ):
            assert log == truth

    def test_help_live(self) -> None:
        """Test `cobib.tui.tui.TUI.help` while the TUI is actually running."""

        def assertion(screen, logs, **kwargs):  # type: ignore
            # pylint: disable=consider-using-f-string
            header_check = ["coBib TUI Help" in line for line in screen.display]
            assert any(header_check)
            offset = header_check.index(True)
            for cmd, desc in TUI.HELP_DICT.items():
                assert any(
                    "{:<8} {}".format(cmd + ":", desc) in line
                    for line in screen.display[2 + offset : 19 + offset]
                )

        self.run_tui("?", assertion, {})

    def test_help_no_artifacts(self) -> None:
        """Test that the help popup leaves no rendering artifacts behind."""

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_lines = [
                "knuthwebsite    Knuth: Computers and Typesetting",
                "latexcompanion  The \\LaTeX\\ Companion",
                'einstein        Zur Elektrodynamik bewegter K{\\"o}rper',
            ]
            for line, truth in zip_longest(expected_lines, screen.display[1:4]):
                assert line == truth.strip()
            for line in range(5, 22):
                assert screen.display[line].strip() == ""

        self.run_tui("?q", assertion, {})

    @pytest.mark.parametrize(
        ["selection"],
        [
            [set()],
            [{"knuthwebsite"}],
        ],
    )
    def test_select(
        self, patch_curses: Any, caplog: pytest.LogCaptureFixture, selection: Set[str]
    ) -> None:
        """Test `cobib.tui.tui.TUI.select`.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
            selection: the set of selected labels.
        """
        stdscr = MockCursesPad()
        stdscr.size = (24, 80)
        tui = TUI(stdscr, debug=True)
        tui.selection = copy.deepcopy(selection)
        caplog.clear()

        tui.select()
        assert tui.selection == (set() if selection else {"knuthwebsite"})
        expected_log = [
            ("cobib.tui.tui", 10, "Select command triggered."),
            ("cobib.tui.frame", 10, 'Obtaining current label "under" cursor.'),
            ("cobib.tui.frame", 10, 'Current label at "0" is "knuthwebsite".'),
        ]
        if selection:
            expected_log.append(
                ("cobib.tui.tui", 20, "Removing 'knuthwebsite' from the selection.")
            )
        else:
            expected_log.append(("cobib.tui.tui", 20, "Adding 'knuthwebsite' to the selection."))
        assert [
            record
            for record in caplog.record_tuples
            if record[0] in ("cobib.tui.frame", "cobib.tui.tui")
        ] == expected_log

    @pytest.mark.parametrize(
        ["keys", "assertion_kwargs"],
        [
            ["v", {"current": 1, "selected": [1], "labels": ["knuthwebsite"]}],
            ["vj", {"current": 2, "selected": [1], "labels": ["knuthwebsite"]}],
            ["jvjv", {"current": 3, "selected": [2, 3], "labels": ["latexcompanion", "einstein"]}],
            ["vv", {"current": 1, "selected": [], "labels": []}],
            # ['v\n', assert_select_show_view, {'current': True}],
        ],
    )
    def test_select_list_live(self, keys: str, assertion_kwargs: Dict[str, Any]) -> None:
        """Tests the select method in the list view while the TUI is actually running.

        Args:
            keys: the string of keys to send to the running TUI.
            assertion_kwargs: additional keyword arguments to pass to the assertion method.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            term_width = len(screen.buffer[0])
            current = kwargs["current"]
            labels = kwargs["labels"]
            selected = kwargs["selected"]
            for sel, lab in zip(selected, labels):
                assert [c.bg for c in screen.buffer[sel].values()] == ["magenta"] * len(lab) + [
                    "cyan" if sel == current else "default"
                ] * (term_width - len(lab))
            if not selected and not labels:
                for idx in range(1, 4):
                    assert [c.bg for c in screen.buffer[idx].values()] == [
                        "cyan" if idx == current else "default"
                    ] * term_width

        self.run_tui(keys, assertion, assertion_kwargs)

    def test_select_show_live(self) -> None:
        """Tests the select method in the show view while the TUI is actually running."""

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_screen = [
                "@misc{knuthwebsite,",
                " author = {Donald Knuth},",
                " title = {Knuth: Computers and Typesetting},",
                r" url = {http://www-cs-faculty.stanford.edu/\~{}uno/abcde.html}",
                "}",
            ]
            term_width = len(screen.buffer[0])
            label_len = len("knuthwebsite")
            for idx, (line, truth) in enumerate(zip(screen.display[1:6], expected_screen)):
                if idx == 0:
                    assert [c.bg for c in screen.buffer[1].values()] == ["cyan"] * 6 + [
                        "magenta"
                    ] * label_len + ["cyan"] * (term_width - label_len - 6)
                assert line.strip() in truth.strip()

        self.run_tui("v\n", assertion, {})

    def test_select_search_live(self) -> None:
        """Tests the select method in the search view while the TUI is actually running."""

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_screen = [
                "knuthwebsite - 1 match",
                "[1]     @misc{knuthwebsite,",
                "[1]      author = {Donald Knuth},",
            ]
            term_width = len(screen.buffer[0])
            label_len = len("knuthwebsite")
            for idx, (line, truth) in enumerate(zip(screen.display[1:6], expected_screen)):
                if idx == 0:
                    assert [c.bg for c in screen.buffer[1].values()] == ["magenta"] * label_len + [
                        "cyan"
                    ] * (term_width - label_len)
                assert line.strip() in truth.strip()

        self.run_tui("v/knuth\n", assertion, {})

    @pytest.mark.parametrize(
        ["text"],
        [
            [["Some single-line dummy text."]],
            [["Some multi-line dummy text.", "This is the second line."]],
        ],
    )
    def test_prompt_print(
        self, patch_curses: Any, caplog: pytest.LogCaptureFixture, text: List[str]
    ) -> None:
        """Test `cobib.tui.tui.TUI.prompt_print`.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
            text: the text to print to the prompt.
        """
        stdscr = MockCursesPad()
        stdscr.size = (24, 80)
        tui = TUI(stdscr, debug=True)
        caplog.clear()

        tui.prompt_print("\n".join(text))
        assert tui.prompt.lines == [text[0]]  # type: ignore
        if len(text) > 1:
            # assert popup on multi-line text messages
            assert (
                "cobib.tui.buffer",
                10,
                "Appending string to text buffer: " + "\n".join(text),
            ) in caplog.record_tuples
            assert ("cobib.tui.buffer", 10, "Create popup window.") in caplog.record_tuples

    @pytest.mark.parametrize(
        ["prompt_quit", "returned_char", "mode"],
        [
            [False, 27, Mode.LIST.value],
            [False, 27, Mode.SHOW.value],
            [False, 27, Mode.SEARCH.value],
            [True, ord("y"), Mode.LIST.value],
            [True, ord("n"), Mode.LIST.value],
        ],
    )
    def test_quit(
        self,
        patch_curses: Any,
        caplog: pytest.LogCaptureFixture,
        prompt_quit: bool,
        returned_char: int,
        mode: str,
    ) -> None:
        """Test `cobib.tui.tui.TUI.quit`.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
            prompt_quit: whether to prompt before actually quitting.
            returned_char: the value for `tests.tui.mock_curses.MockCursesPad.returned_chars`.
            mode: the `cobib.tui.state.Mode` value.
        """
        stdscr = MockCursesPad()
        stdscr.size = (24, 80)
        config.tui.prompt_before_quit = prompt_quit
        tui = TUI(stdscr, debug=True)
        STATE.mode = mode
        caplog.clear()

        tui.prompt.returned_chars = [returned_char]  # type: ignore

        expected_log = []
        if mode == Mode.LIST.value:
            expected_log.append(("cobib.tui.tui", 10, "Quitting from lowest level."))
        else:
            expected_log.append(
                ("cobib.tui.tui", 10, "Quitting higher menu level. Falling back to list view.")
            )
        if prompt_quit:
            expected_log.append(("TUITest", 10, "curs_set: (1,)"))

        if returned_char == ord("n"):
            expected_log.append(("cobib.tui.tui", 20, "User aborted quitting."))
            expected_log.append(("TUITest", 10, "curs_set: (0,)"))

        if mode == Mode.LIST.value and returned_char != ord("n"):
            with pytest.raises(StopIteration):
                tui.quit()
        else:
            tui.quit()
        assert [
            record for record in caplog.record_tuples if record[0] in ("cobib.tui.tui", "TUITest")
        ] == expected_log

    # Most keys are difficult to test here and we assume that the breadth of this function is tested
    # by the unittests of the triggered commands. Nonetheless, we test a few special keys to ensure
    # that the method itself works roughly as intended.
    @pytest.mark.parametrize(
        ["keys"],
        [[[ord("q")]], [[ord("q"), 410]]],  # curses.KEY_RESIZE
    )
    def test_loop(
        self,
        patch_curses: Any,
        caplog: pytest.LogCaptureFixture,
        keys: Union[str, List[List[int]]],
    ) -> None:
        """Test `cobib.tui.tui.TUI.loop`.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
            keys: the keys to send to the TUI.
        """
        stdscr = MockCursesPad()
        stdscr.size = (24, 80)
        stdscr.returned_chars = keys  # type: ignore
        config.tui.prompt_before_quit = False
        tui = TUI(stdscr, debug=True)
        # we expect normal execution
        tui.loop(debug=True)
        # minimal assertions
        for key in keys:
            assert ("cobib.tui.tui", 10, f"Key press registered: {key}") in caplog.record_tuples
        assert caplog.record_tuples[-2] == ("cobib.tui.tui", 10, "Quitting from lowest level.")
        assert caplog.record_tuples[-1] == ("cobib.tui.tui", 10, "Stopping key event loop.")

    def test_loop_inactive_commands(
        self, patch_curses: Any, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test that inactive commands are not triggered during the TUI loop.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            caplog: the built-in pytest fixture.
        """
        stdscr = MockCursesPad()
        stdscr.size = (24, 80)
        stdscr.returned_chars = [ord("q"), ord("\n")]
        config.tui.prompt_before_quit = False
        tui = TUI(stdscr, debug=True)
        tui.STATE.inactive_commands = ["Show"]
        # we expect normal execution
        tui.loop(debug=True)
        # minimal assertions
        assert (
            "cobib.commands.show",
            10,
            "Show command triggered from TUI.",
        ) not in caplog.record_tuples
        for key in stdscr.returned_chars:
            assert ("cobib.tui.tui", 10, f"Key press registered: {key}") in caplog.record_tuples
        assert caplog.record_tuples[-2] == ("cobib.tui.tui", 10, "Quitting from lowest level.")
        assert caplog.record_tuples[-1] == ("cobib.tui.tui", 10, "Stopping key event loop.")

    @pytest.mark.parametrize(
        ["keys", "expected"],
        [
            [["t", "e", "s", "t", "\n"], "test"],  # normal Enter
            [["t", "e", "s", "t", 27, None, -1], ""],  # Escape
            [["t", "e", 27, None, 68, "s", "t", "\n"], "tst"],  # left arrow
            [
                ["t", "e", 27, None, 68, 27, None, 67, "s", "t", "\n"],
                "test",
            ],  # left and right arrow
            [["t", "e", 127, "s", "t", "\n"], "tst"],  # internal Backspace
            [[127, 127], ""],  # Backspace until prompt is empty
        ],
    )
    def test_prompt_handler(
        self, patch_curses: Any, keys: List[Union[int, str]], expected: str
    ) -> None:
        """Test `cobib.tui.tui.TUI.prompt_handler`.

        Args:
            patch_curses: the `tests.tui.tui_test.TUITest.patch_curses` fixture.
            keys: the keys to send to the prompt handler.
            expected: the expected string to be returned from the prompt handler.
        """
        stdscr = MockCursesPad()
        stdscr.size = (24, 80)
        config.tui.prompt_before_quit = False
        tui = TUI(stdscr, debug=True)
        tui.prompt.returned_chars = [  # type: ignore
            ord(k) if isinstance(k, str) else k for k in reversed(keys)
        ]
        command = tui.prompt_handler("")
        assert command == expected

    @pytest.mark.parametrize(
        ["keys", "assertion_kwargs"],
        [
            ["x", {"contents": ":export"}],
            [":", {"contents": ":"}],
            [":" + 100 * "j", {"contents": "j" * 79}],  # regression-test against #48
        ],
    )
    def test_prompt_live(self, keys: str, assertion_kwargs: Dict[str, str]) -> None:
        """Test `cobib.tui.tui.TUI.prompt_handler` while the TUI is actually running.

        Args:
            keys: the string of keys to send to the running TUI.
            assertion_kwargs: additional keyword arguments to pass to the assertion method.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            assert screen.display[-1].strip() == kwargs["contents"]

        self.run_tui(keys, assertion, assertion_kwargs)

    def test_execute_command(self) -> None:
        """Test `cobib.tui.tui.TUI.execute_command`."""
        pytest.skip("This method is tested implicitly by the test.commands.*.test_tui methods.")

    @pytest.mark.parametrize(
        ["keys", "assertion_kwargs"],
        [
            # vertical scrolling
            ["G", {"update": 20, "direction": "y"}],
            ["Gg", {"update": 0, "direction": "y"}],
            ["j", {"update": 1, "direction": "y"}],
            ["jjk", {"update": 1, "direction": "y"}],
            # assert scrolloff value of `3` is respected
            ["".join(["j"] * 20), {"update": 17, "direction": "y"}],
            ["".join(["j"] * 21), {"update": 18, "direction": "y"}],
            ["".join(["j"] * 22), {"update": 19, "direction": "y"}],
            ["G" + "".join(["k"] * 20), {"update": 3, "direction": "y"}],
            ["G" + "".join(["k"] * 21), {"update": 2, "direction": "y"}],
            ["G" + "".join(["k"] * 22), {"update": 1, "direction": "y"}],
            # horizontal scrolling
            ["l", {"update": 1, "direction": "x"}],
            ["llh", {"update": 1, "direction": "x"}],
            ["$", {"update": 23, "direction": "x"}],
            ["$0", {"update": 0, "direction": "x"}],
        ],
    )
    def test_scroll_live(self, keys: str, assertion_kwargs: Dict[str, Union[int, str]]) -> None:
        """Test scrolling while the TUI is actually running.

        Args:
            keys: the string of keys to send to the running TUI.
            assertion_kwargs: additional keyword arguments to pass to the assertion method.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            direction = kwargs["direction"]
            update = kwargs["update"]
            term_width = len(screen.buffer[0])

            if direction == "y" or update == 0:
                assert [c.fg for c in screen.buffer[1 + update].values()] == ["white"] * term_width
                assert [c.bg for c in screen.buffer[1 + update].values()] == ["cyan"] * term_width
            elif direction == "x":
                # TODO: figure out how (or if) to actually use the update information
                assert [c.fg for c in screen.buffer[1].values()] == ["white"] * term_width
                assert [c.bg for c in screen.buffer[1].values()] == ["cyan"] * term_width

        # overwrite database
        config.database.file = get_resource("scrolling_database.yaml", "tui")
        try:
            Database().read()
            self.run_tui(keys, assertion, assertion_kwargs)
        finally:
            Database().clear()

    @pytest.mark.parametrize(
        ["keys", "assertion_kwargs"],
        [
            ["w", {"state": True}],
            ["ww", {"state": False}],
        ],
    )
    def test_wrap_live(self, keys: str, assertion_kwargs: Dict[str, bool]) -> None:
        """Test wrapping while the TUI is actually running.

        Args:
            keys: the string of keys to send to the running TUI.
            assertion_kwargs: additional keyword arguments to pass to the assertion method.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            state = kwargs["state"]
            if state:
                assert screen.display[2][:2] == TextBuffer.INDENT + " "
            else:
                assert screen.display[1][:2] == "kn"

        self.run_tui([SIGWINCH] + list(keys), assertion, assertion_kwargs)  # type: ignore

    @pytest.mark.parametrize(
        ["kwargs"],
        [
            [[]],
            [["-l", str(PurePath(tempfile.gettempdir()) / "cobib_test.log")]],
        ],
    )
    def test_cmdline(self, monkeypatch: pytest.MonkeyPatch, kwargs: List[str]) -> None:
        """Test the command-line access of the TUI.

        Args:
            monkeypatch: the built-in pytest fixture.
            kwargs: the additional arguments to pass to the command line call.
        """
        monkeypatch.setattr("cobib.tui.tui", lambda: None)
        self.run_module(monkeypatch, "main", ["cobib"] + kwargs)



#!/usr/bin/env python3
"""coBib main body."""

import argparse
import inspect
import logging
import platform
import sys

from cobib import __version__, commands
from cobib.config import config
from cobib.database import Database
from cobib.utils import shell_helper
from cobib.utils.logging import get_file_handler, get_stream_handler, print_changelog

LOGGER = logging.getLogger(__name__)


def main() -> None:
    """Main executable.

    coBib's main function used to parse optional keyword arguments and subcommands.
    """
    if len(sys.argv) > 1 and any(a[0] == "_" for a in sys.argv):
        # shell helper function called
        helper_main()
        sys.exit()

    # initialize logging
    root_logger = logging.getLogger()
    root_logger.setLevel("DEBUG")
    stream_handler = get_stream_handler()
    root_logger.addHandler(stream_handler)

    subcommands = [cmd.split(":")[0] for cmd in shell_helper.list_commands([])]
    parser = argparse.ArgumentParser(
        prog="coBib",
        description=(
            "Cobib input arguments.\nIf no arguments are given, the TUI will start as a default."
        ),
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s v{__version__}")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("-l", "--logfile", type=argparse.FileType("w"), help="Alternative log file")
    parser.add_argument(
        "-c", "--config", type=argparse.FileType("r"), help="Alternative config file"
    )
    parser.add_argument("command", help="subcommand to be called", choices=subcommands, nargs="?")
    parser.add_argument("args", nargs=argparse.REMAINDER)

    args = parser.parse_args()

    if args.logfile:
        LOGGER.info("Switching to FileHandler logger in %s", args.logfile.name)
        file_handler = get_file_handler(
            "DEBUG" if args.verbose > 1 else "INFO", logfile=args.logfile.name
        )
        root_logger.addHandler(file_handler)

    # set logging verbosity level
    if args.verbose == 1:
        stream_handler.setLevel(logging.INFO)
        LOGGER.info("Logging level set to INFO.")
    elif args.verbose > 1:
        stream_handler.setLevel(logging.DEBUG)
        LOGGER.info("Logging level set to DEBUG.")

    # load configuration
    config.load(args.config)

    # print latest changelog
    print_changelog(__version__, config.logging.version)

    if args.command == "init":
        # the database file may not exist yet, thus we ensure to execute the command before trying
        # to read the database file
        subcmd = getattr(commands, "InitCommand")()
        subcmd.execute(args.args)
        return

    # initialize database
    Database()

    if not args.command:
        if platform.system() not in ("Linux", "Darwin"):
            LOGGER.error("Only Linux and Darwin systems are supported by coBib's TUI!")
            return

        # pylint: disable=import-outside-toplevel
        from cobib.tui import tui

        if args.logfile is None:
            LOGGER.info('Switching to FileHandler logger in "%s"', config.logging.logfile)
            file_handler = get_file_handler("DEBUG" if args.verbose > 1 else "INFO")
            root_logger.addHandler(file_handler)
        else:
            LOGGER.info(
                'Already logging to "%s". NOT switching to "%s"',
                args.logfile,
                config.logging.logfile,
            )

        stream_handler.setLevel(logging.WARNING)

        tui()
    else:
        subcmd = getattr(commands, args.command.title() + "Command")()
        subcmd.execute(args.args)


def helper_main() -> None:
    """Shell helper.

    This is an auxiliary main method which exposes some shell utility methods.
    """
    # initialize logging
    root_logger = logging.getLogger()
    root_logger.setLevel("DEBUG")
    stream_handler = get_stream_handler()
    root_logger.addHandler(stream_handler)

    available_helpers = [
        "_" + m[0] for m in inspect.getmembers(shell_helper) if inspect.isfunction(m[1])
    ]
    parser = argparse.ArgumentParser(description="Process shell helper call")
    parser.add_argument("helper", help="shell helper to be called", choices=available_helpers)
    parser.add_argument("args", nargs=argparse.REMAINDER)
    parser.add_argument("--verbose", "-v", action="count", default=0)
    parser.add_argument("-l", "--logfile", type=argparse.FileType("w"), help="Alternative log file")
    parser.add_argument(
        "-c", "--config", type=argparse.FileType("r"), help="Alternative config file"
    )

    args = parser.parse_args()

    if args.logfile:
        LOGGER.info("Switching to FileHandler logger in %s", args.logfile.name)
        file_handler = get_file_handler(
            "DEBUG" if args.verbose > 1 else "INFO", logfile=args.logfile.name
        )
        root_logger.addHandler(file_handler)

    # set logging verbosity level
    if args.verbose == 1:
        stream_handler.setLevel(logging.INFO)
        LOGGER.info("Logging level set to INFO.")
    elif args.verbose > 1:
        stream_handler.setLevel(logging.DEBUG)
        LOGGER.info("Logging level set to DEBUG.")

    # load configuration
    config.load(args.config)

    helper = getattr(shell_helper, args.helper.strip("_"))
    # any shell helper function will return a list of the requested items
    for item in helper(args.args):
        print(item)


if __name__ == "__main__":
    main()  # pragma: no cover



"""coBib's Redo command.

This command can be used to re-apply the changes *of a previously undone* command:
```
cobib redo
```
This command takes *no* additional arguments!

Note, that if you have not used `cobib undo` previously, this command will have no effect!

Furthermore, this command is *only* available if coBib's git-integration has been enabled and
initialized.
Refer to the documentation of `cobib.commands.init.InitCommand` for more details on that topic.

You can also trigger this command from the `cobib.tui.tui.TUI`.
By default, it is bound to the `r` key.
"""

from __future__ import annotations

import argparse
import logging
import subprocess
import sys
from typing import IO, TYPE_CHECKING, Any, List

from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class RedoCommand(Command):
    """The Redo Command."""

    name = "redo"

    def execute(self, args: List[str], out: IO[Any] = sys.stdout) -> None:
        """Redoes the last undone change.

        This command is *only* available if coBib's git-integration has been enabled via
        `config.database.git` *and* initialized properly (see `cobib.commands.init.InitCommand`).
        If that is the case, this command will re-apply the changes *of a previously undone* command
        (see `cobib.commands.undo.UndoCommand`).

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * **no** additional arguments are required for this subcommand!
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        git_tracked = config.database.git
        if not git_tracked:
            msg = (
                "You must enable coBib's git-tracking in order to use the `Redo` command."
                "\nPlease refer to the man-page for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        file = RelPath(config.database.file).path
        root = file.parent
        if not (root / ".git").exists():
            msg = (
                "You have configured, but not initialized coBib's git-tracking."
                "\nPlease consult `cobib init --help` for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        LOGGER.debug("Starting Redo command.")
        parser = ArgumentParser(prog="redo", description="Redo subcommand parser.")

        try:
            # pylint: disable=unused-variable
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return

        Event.PreRedoCommand.fire(largs)

        LOGGER.debug("Obtaining git log.")
        lines = subprocess.check_output(
            [
                "git",
                "--no-pager",
                "-C",
                f"{root}",
                "log",
                "--oneline",
                "--no-decorate",
                "--no-abbrev",
            ]
        )
        redone_shas = set()
        for commit in lines.decode().strip().split("\n"):
            LOGGER.debug("Processing commit %s", commit)
            sha, *message = commit.split()
            if message[0] == "Redo":
                # Store already redone commit sha
                LOGGER.debug("Storing redone commit sha: %s", message[-1])
                redone_shas.add(message[-1])
                continue
            if sha in redone_shas:
                LOGGER.info("Skipping %s as it was already redone", sha)
                continue
            if message[0] == "Undo":
                LOGGER.debug("Attempting to redo %s.", sha)
                commands = [
                    f"git -C {root} revert --no-commit {sha}",
                    f"git -C {root} commit --no-gpg-sign --quiet --message 'Redo {sha}'",
                ]
                with subprocess.Popen(
                    "; ".join(commands), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                ) as redo:
                    redo.communicate()
                    if redo.returncode != 0:
                        LOGGER.error(  # pragma: no cover
                            "Redo was unsuccessful. Please consult the logs and git history of your"
                            " database for more information."
                        )
                    else:
                        # update Database
                        Database().read()
                break
        else:
            msg = "Could not find a commit to redo. You must have undone something first!"
            LOGGER.warning(msg)
            sys.exit(1)

        Event.PostRedoCommand.fire(root, sha)

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("Redo command triggered from TUI.")
        tui.execute_command(["redo"], skip_prompt=True)
        # update database list
        LOGGER.debug("Updating list after Redo command.")
        tui.viewport.update_list()



"""Tests for coBib's logging helper functions."""

import logging
import re
import tempfile
from typing import Optional

import pytest

from cobib.utils.logging import get_file_handler, get_stream_handler, print_changelog

from .. import MockStdin, get_resource


def test_get_stream_handler() -> None:
    """Test stream logging configuration."""
    handler = get_stream_handler()
    assert handler.level == 30
    assert isinstance(handler, logging.StreamHandler)


def test_get_file_handler() -> None:
    """Test file logging configuration."""
    handler = get_file_handler("INFO")
    assert handler.level == 20
    assert isinstance(handler, logging.FileHandler)


@pytest.mark.parametrize("cached_version", [None, False, "", "0.1", "0.2"])
def test_print_changelog(
    cached_version: Optional[str],
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """Test printing of the latest changelog section.

    Args:
        cached_version: the cached version to test against.
        monkeypatch: the built-in pytest fixture.
        capsys: the built-in pytest fixture.
    """
    monkeypatch.setattr("sys.stdin", MockStdin())
    ansi_regex = re.compile(r"(\x1b\[(\d+)+m)")
    if cached_version is None:
        print_changelog("0.2", None)
        assert capsys.readouterr().out == ""
    else:
        with tempfile.NamedTemporaryFile("w") as file:
            if not cached_version:
                file.close()
            else:
                file.write(cached_version)
                file.flush()
            print_changelog("0.2", file.name)
        captured = ansi_regex.sub("", capsys.readouterr().out).splitlines()
        if cached_version == "0.2":
            assert captured == []
        with open(
            get_resource("expected_changelog_printing.txt", "utils"), "r", encoding="utf-8"
        ) as expected:
            expected_lines = expected.read().splitlines()
            for true, exp in zip(captured[:-2], expected_lines):
                assert true == exp
            for true, exp in zip(captured[-2:], expected_lines[-2:]):
                assert true == exp
